"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""
import subprocess
import numpy as np
import h5py
import os
from numpy.lib.recfunctions import drop_fields, append_fields
from functools import partial

from picopore.util import log, isGroup, getDtype, findDatasets, rewriteDataset, recursiveCollapseGroups, uncollapseGroups, getPrefixedFilename

__basegroup_name__ = "Picopore"
__raw_compress_keywords__ = ["Alignment","Log","Configuration","HairpinAlign","Calibration_Strand","Hairpin_Split","EventDetection","Events","Segmentation"]
__raw_compress_summary__ = ["Summary"]
__raw_compress_fastq__ = ["BaseCalled"]
__raw_compress_fastq_summary__ = ["Basecall"]

def chooseCompressFunc(revert, mode, fastq, summary, manual, realtime=False):
    name = "Performing "
    if realtime:
        name += "real time "
    if revert:
        if mode == 'lossless':
            func = losslessDecompress
            name += "lossless decompression"
        elif mode == 'deep-lossless':
            func = deepLosslessDecompress
            name += "deep lossless decompression"
        else:
            log("Unable to revert raw files. Please use a basecaller instead.")
            exit(1)
    else:
        if mode == 'lossless':
            func = losslessCompress
            name += "lossless compression"
        elif mode == 'deep-lossless':
            func = deepLosslessCompress
            name += "deep lossless compression"
        elif mode == 'raw':
            name += "raw compression "
            if manual is not None:
                name += "with manual keyword " + manual
                keywords = [manual]
            else:
                keywords = __raw_compress_keywords__
                if fastq and summary:
                    name += "with FASTQ and summary"
                elif fastq:
                    keywords += __raw_compress_summary__
                    name += "with FASTQ and no summary"
                elif summary:
                    keywords += __raw_compress_fastq__
                    name += "with summary and no FASTQ"
                else:
                    keywords += __raw_compress_fastq_summary__
                    name += "with no summary and no FASTQ"
            func = partial(rawCompress, keywords=keywords)
    try:
        return partial(compress, func), name
    except NameError:
        log("No compression method selected")
        exit(1)

def indexToZero(f, path, col, name="picopore.{}_index", dataColumn=None):
    dataset = f[path]
    name = name.format(col)
    data = f[path].value
    if not name in dataset.attrs.keys():
        dataColumn = data[col] if dataColumn is None else dataColumn
        start_index = min(dataColumn)
        dataset.attrs.create(name, start_index, dtype=getDtype(start_index))
        dataColumn = dataColumn-start_index
        data = drop_fields(data, [col])
        data = append_fields(data, [col], [dataColumn], [getDtype(dataColumn)])
    return data

def deepLosslessCompress(f, group):
    paths = findDatasets(f, group, "Events")
    paths = [path for path in paths if "Basecall" in path]
    # index event detection
    if "UniqueGlobalKey/channel_id" in f:
        sampleRate = f["UniqueGlobalKey/channel_id"].attrs["sampling_rate"]
        for path in paths:
            if f[path].parent.parent.attrs.__contains__("event_detection"):
                # index back to event detection
                dataset = f[path].value
                start = np.array([int(round(sampleRate * i)) for i in dataset["start"]])
                dataset = indexToZero(f, path, "start", dataColumn=start)
                move = dataset["move"] # rewrite move dataset because it's int64 for max 2
                # otherwise, event by event
                dataset = drop_fields(dataset, ["mean", "stdv", "length", "move"])
                dataset = append_fields(dataset, ["move"], [move], [getDtype(move)])
                rewriteDataset(f, path, compression="gzip", compression_opts=9, dataset=dataset)
                # rewrite eventdetection too - start is also way too big here
                eventDetectionPath = findDatasets(f, "all", entry_point=f[path].parent.parent.attrs.get("event_detection"))[0]
                if "picopore.start_index" not in f[eventDetectionPath].attrs.keys():
                    eventData = indexToZero(f, eventDetectionPath, "start")
                    rewriteDataset(f, eventDetectionPath, compression="gzip", compression_opts=9, dataset=eventData)

    if __basegroup_name__ not in f:
        f.create_group(__basegroup_name__)
        for name, group in f.items():
            if name != __basegroup_name__:
                recursiveCollapseGroups(f, __basegroup_name__, name, group)
    return losslessCompress(f, group)

def deepLosslessDecompress(f, group):
    # rebuild group hierarchy
    if __basegroup_name__ in f.keys():
        uncollapseGroups(f, f[__basegroup_name__])
    paths = findDatasets(f, group)
    paths = [path for path in paths if "Basecall" in path]
    sampleRate = f["UniqueGlobalKey/channel_id"].attrs["sampling_rate"]
    for path in paths:
        if f[path].parent.parent.attrs.__contains__("event_detection"):
            # index back to event detection
            dataset = f[path].value
            if "mean" not in dataset.dtype.names:
                eventDetectionPath = findDatasets(f, "all", entry_point=f[path].parent.parent.attrs.get("event_detection"))[0]
                eventData = f[eventDetectionPath].value
                try:
                    start = eventData["start"] + f[eventDetectionPath].attrs["picopore.start_index"]
                    del f[eventDetectionPath].attrs["picopore.start_index"]
                    eventData = drop_fields(eventData, ["start"])
                    eventData = append_fields(eventData, ["start"], [start], [getDtype(start)])
                    rewriteDataset(f, eventDetectionPath, compression="gzip", compression_opts=1, dataset=eventData)
                except KeyError:
                    # must have been compressed without start indexing
                    pass
                try:
                    start_index = f[path].attrs["picopore.start_index"]
                    del f[path].attrs["picopore.start_index"]
                except KeyError:
                    # must have been compressed without start indexing
                    start_index=0
                start = dataset["start"][0] + start_index
                end = dataset["start"][-1] + start_index
                # constrain to range in basecall
                eventData = eventData[np.logical_and(eventData["start"] >= start, eventData["start"] <= end)]
                # remove missing events
                i=0
                keepIndex = []
                for time in dataset["start"]:
                    while eventData["start"][i] != time + start_index and i < eventData.shape[0]:
                        i += 1
                    keepIndex.append(i)
                eventData = eventData[keepIndex]
                dataset = drop_fields(dataset, "start")
                start = [i/sampleRate for i in eventData["start"]]
                length = [i/sampleRate for i in eventData["length"]]
                dataset = append_fields(dataset, ["mean", "start", "stdv", "length"], [eventData["mean"], start, eventData["stdv"], length])
                rewriteDataset(f, path, dataset=dataset)
    return losslessDecompress(f, group)

def losslessCompress(f, group):
    paths = findDatasets(f, group, keyword="Events")
    paths.extend(findDatasets(f, group, keyword="Alignment"))
    paths.extend(findDatasets(f, "all", keyword="Signal", entry_point="Raw"))
    for path in paths:
        rewriteDataset(f, path, "gzip", 9)
    return "GZIP=9"

def losslessDecompress(f, group):
    paths = findDatasets(f, group, keyword="Events")
    paths.extend(findDatasets(f, group, keyword="Alignment"))
    paths.extend(findDatasets(f, "all", keyword="Signal", entry_point="Raw"))
    for path in paths:
        rewriteDataset(f, path)
    return "GZIP=1"

def rawCompress(f, group, keywords):
    if "Picopore" in f:
        log("{} is compressed using picopore deep-lossless compression. Please use picpore --revert --mode deep-lossless before attempting raw compression.".format(f.filename))
    else:
        paths = []
        for kw in keywords:
            paths.extend(findDatasets(f, group, keyword=kw))
        for path in paths:
            if path in f:
                del f[path]
        try:
            if len(f["Analyses"].keys()) == 0:
                del f["Analyses"]
        except KeyError:
            # no analyses, no worries
            pass
    return "GZIP=9"

def compress(func, filename, group="all"):
    try:
        with h5py.File(filename, 'r+') as f:
            filtr = func(f, group)
        subprocess.call(["h5repack","-f",filtr,filename, "{}.tmp".format(filename)])
        subprocess.call(["mv","{}.tmp".format(filename),filename])
        return os.path.getsize(filename)
    except Exception as e:
        log("ERROR: {} on file {}".format(str(e), filename))
        if os.path.isfile("{}.tmp".format(filename)):
            os.remove("{}.tmp".format(filename))
        return os.path.getsize(filename)



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

# Original code at https://github.com/jreese/multiprocessing-keyboardinterrupt
# Copyright (c) 2011 John Reese
# Licensed under the MIT License

import multiprocessing
import signal

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

class Multiprocessor:

    def __init__(self, threads):
        self.pool = self.init_pool(threads)
        self.results = []

    def init_pool(self, threads):
        return multiprocessing.Pool(threads, init_worker)

    def apply_async(self, func, argList):
        for arg in argList:
            self.results.append(self.pool.apply_async(func, args=[arg]))

    def join(self):
        try:
            return [r.get() for r in self.results]

        except KeyboardInterrupt:
            self.pool.terminate()
            self.pool.join()
            raise

    def wait(self):
        try:
            for r in self.results:
                r.wait()
            return 0

        except KeyboardInterrupt:
            self.pool.terminate()
            self.pool.join()
            raise

    def __del__(self):
        self.pool.close()
        self.pool.join()



from __future__ import print_function
import subprocess
import os
import errno
import shutil
import time
import signal

__test_files__ = ["sample_data/albacore_1d_original.fast5", "sample_data/metrichor_2d_original.fast5"]
__test_runs__ = ["lossless", "deep-lossless"]
__prefix__ = "picopore.test"
__raw_runs__ = [["--fastq","--summary"],["--summary","--no-fastq"],["--fastq","--no-summary"],["--no-fastq","--no-summary"], ["--manual", "Analyses"]]

def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def call(additionalArgs, prefix=None):
    args=["python","-m","picopore","-y","--print-every","1"]
    if prefix is not None:
        args.extend(["--prefix",prefix])
    args.extend(additionalArgs)
    print(" ".join(args))
    p = subprocess.call(args)
    if prefix is not None:
        filename = args[-1]
        dirname = os.path.dirname(filename)
        basename = os.path.basename(filename)
        os.remove(os.path.join(dirname,".".join([prefix,basename])))
    return p

def testFile(filename):
    result = 0
    for run in __test_runs__:
        result += call(["--test","--mode",run, filename])
        result += call(["--mode",run, filename], prefix=__prefix__)
    for run in __raw_runs__:
        result += call(["--mode","raw"] + run + [filename], prefix=__prefix__)
    return result

def testRealtime(mode, additionalArgs=None, directory="realtime"):
    __waittime = 10
    mkdir(directory)
    args = ["python","-m","picopore","-y","--realtime","--print-every","1"]
    if additionalArgs is not None:
        args.extend(additionalArgs)
    args.extend(["--mode",mode,directory])
    print(" ".join(args))
    p = subprocess.Popen(args)
    time.sleep(__waittime)
    for filename in __test_files__:
        shutil.copy(filename, directory)
        time.sleep(__waittime)
    p.send_signal(signal.SIGINT)
    p.wait()
    shutil.rmtree(directory)
    return p.returncode

exitcode = 0
for filename in __test_files__:
    exitcode += testFile(filename)

for mode in __test_runs__:
    exitcode += testRealtime(mode)
for mode in __raw_runs__:
    exitcode += testRealtime("raw", additionalArgs=mode)
exit(exitcode)



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import os

from picopore.util import log, isGroup, isArray, getPrefixedFilename
from picopore.runner import PicoporeCompressionRunner
from picopore.parse_args import parseArgs

def checkContents(obj1, obj2, name=None):
    name = obj1.name if name is None else name
    keys1 = obj1.keys()
    keys2 = obj2.keys()
    exitcode = 0
    for key in keys1:
        if key not in keys2:
            log("Failure: {} missing from file 2".format("/".join([obj2.name, key])))
            exitcode += 1
    for key in keys2:
        if key not in keys1:
            log("Failure: {} missing from file 1".format("/".join([obj1.name, key])))
            exitcode += 1
    return exitcode

def checkData(data1, data2, name):
    match = data1==data2
    exitcode = 0
    if isArray(match):
        match = (data1==data2).all()
    if not match:
        positions = [i for i in range(len(data1)) if not data1[i] == data2]
        for pos in positions:
            log("Failure: {}[{}] - file1={}, file2={}".format(name, pos, data1[pos], data2[pos]))
            exitcode += 1
    return exitcode

def recursiveCheckEquivalent(file1, file2, name):
    obj1 = file1[name]
    obj2 = file1[name]
    # check attributes
    attr1 = obj1.attrs
    attr2 = obj2.attrs
    attrsName = "/".join([name, "attrs"])
    exitcode = checkContents(attr1, attr2, attrsName)
    for key, value in attr1.items():
        try:
            if not attr2[key] == value:
                log("Failure: {} - file1={}, file2={}".format("/".join([attrsName, key]), value, attr2[key]))
                exitcode += 1
        except ValueError as e:
            # probably a numpy array
            if str(e) == "The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()" and not (attr2[key] == value).all():
                log("Failure: {} - file1={}, file2={}".format("/".join([attrsName, key]), value, attr2[key]))
                exitcode += 1
    # check subgroups / datasets
    if isGroup(obj1):
        exitcode += checkContents(obj1, obj2)
        for key in obj1.keys():
            exitcode += recursiveCheckEquivalent(file1, file2, "/".join([name, key]))
    else:
        if not obj1.shape == obj2.shape:
            log("Failure: {}.shape - file1={}, file2={}".format(name, obj1.shape, obj2.shape))
        if obj1.dtype.names is None:
            # just one column
            exitcode += checkData(obj1, obj2, name)
        else:
            for col in obj1.dtype.names:
                exitcode += checkData(obj1[col], obj2[col], ".".join([name, col]))
    return exitcode

def checkEquivalent(fn1, fn2):
    log("Checking equivalence of {} (file 1) and {} (file 2)...".format(fn1, fn2))
    with h5py.File(fn1, 'r') as file1, h5py.File(fn2, 'r') as file2:
        exitcode = checkContents(file1, file2)
        for group in file1.values():
            exitcode += recursiveCheckEquivalent(file1, file2, group.name)
    log("Complete with {} errors.".format(exitcode))
    return exitcode
    


class PicoporeTestRunner(PicoporeCompressionRunner):

    def __init__(self, args):
        super(PicoporeTestRunner, self).__init__(args)
        if "lossless" not in self.mode:
            raise ArgumentError(test, "{} mode not reversible by Picopore. Test cancelled.".format(self.mode))
        if self.prefix is None:
            self.prefix = "picopore.test"
        self.fileList = super(PicoporeTestRunner, self).getFileList()
        self.y = True
        self.originalFileList = self.fileList

    def getFileList(self):
        return self.fileList

    def getReversionFileList(self):
        return [getPrefixedFilename(f, self.prefix) for f in self.fileList]

    def execute(self):
        exitcode=1
        if len(self.fileList) == 0:
            return 0
        try:
            self.revert = False
            self.run()

            self.preSize = 0
            self.postSize = 0
            self.revert = True
            self.fileList = self.getReversionFileList()
            self.prefix = None
            self.run()
            exitcode = 0
            for i in range(len(self.fileList)):
                exitcode += checkEquivalent(self.originalFileList[i], self.fileList[i])
        except Exception as e:
            log("ERROR: " + str(e))
        finally:
            for f in self.fileList:
                try:
                    os.remove(f)
                except OSError:
                    # file never created
                    pass
        return exitcode

__description = """"picopore-test compresses to temporary files and checks that all datasets and attributes are equal (lossless modes only)"""
        
def main():
    args = parseArgs(description=__description, prog='picopore-test')
    runner = PicoporeTestRunner(args)
    return runner.execute()

if __name__ == "__main__":
    exit(main())





"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import functools
from shutil import copyfile
import numpy as np

from picopore.parse_args import checkSure
from picopore.util import recursiveFindFast5, log, getPrefixedFilename
from picopore.compress import chooseCompressFunc
from picopore.multiprocess import Multiprocessor

def _process_func(filename, func, prefix, print_every):
        if prefix is not None:
            newFilename = getPrefixedFilename(filename, prefix)
            copyfile(filename, newFilename)
        else:
            newFilename = filename
        result = func(newFilename)
        if result is not None and print_every > 0 and np.random.rand() < 1.0/print_every:
            log('.', end='')
        return result

class AbstractPicoporeRunner(object):

    def __init__(self, args):
        self.y = args.y
        self.multiprocessor = Multiprocessor(args.threads)
        self.prefix = args.prefix
        self.skip_root = args.skip_root
        self.print_every = args.print_every
        self.input = args.input

    def get_func(self):
        raise NotImplementedError()
        return func, message

    def preprocess(self, fileList):
        raise NotImplementedError()

    def postprocess(self, results):
        raise NotImplementedError()
        return 0 # exitcode

    def getFileList(self):
        return recursiveFindFast5(self.input, self.skip_root)

    def process(self, fileList):
        self.preprocess(fileList)
        self.multiprocessor.apply_async(self.func, fileList)

    def stop(self):
        results = self.multiprocessor.join()
        log("Complete.")
        return self.postprocess(results)

    def run(self, postprocess=True):
        func, message = self.get_func()
        self.func = functools.partial(_process_func, func=func, prefix=self.prefix, print_every=self.print_every)
        fileList = self.getFileList()
        if len(fileList) == 0:
            return 0
        log("{} on {} files... ".format(message, len(fileList)))
        if self.y or checkSure():
            self.process(fileList)
            if postprocess:
                return self.stop()
            else:
                return self.multiprocessor.wait()
        else:
            log("User cancelled. Exiting.")
            exit(1)

    def execute(self):
        self.run()
        return 0

class PicoporeCompressionRunner(AbstractPicoporeRunner):

    def __init__(self, args):
        super(PicoporeCompressionRunner, self).__init__(args)
        self.mode = args.mode
        self.revert = args.revert
        self.fastq = args.fastq
        self.summary = args.summary
        self.manual = args.manual
        self.group = args.group
        self.preSize = 0
        self.postSize = 0

    def get_func(self):
        func, message = chooseCompressFunc(self.revert, self.mode, self.fastq, self.summary, self.manual)
        func = functools.partial(func, group=self.group)
        return func, message

    def preprocess(self, fileList):
        self.preSize += sum([os.path.getsize(f) for f in fileList])

    def postprocess(self, results):
        self.postSize = sum(results)
        if self.revert:
            preStr, postStr = "Compressed size:", "Reverted size:"
        else:
            preStr, postStr = "Original size:", "Compressed size:"
        str_len = max(len(preStr), len(postStr)) + 1
        num_len = len(str(max(self.preSize, self.postSize)))
        log("{}{}".format(preStr.ljust(str_len), str(self.preSize).rjust(num_len)))
        log("{}{}".format(postStr.ljust(str_len), str(self.postSize).rjust(num_len)))
        return self.preSize


"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import functools
import h5py

from picopore.parse_args import parseRenameArgs
from picopore.runner import AbstractPicoporeRunner
from picopore.util import findDatasets, log

def rename(filename, pattern, replacement):
    try:
        with h5py.File(filename, 'r+') as f:
            for path in findDatasets(f, entry_point="/", keyword=pattern, match_child=True):
                newPath = re.sub(pattern, replacement, path)
                f[newPath] = f[path]
                del f[path]
                log("Renamed {} to {}".format(path, newPath))
        return 0
    except Exception as e:
        log("ERROR: {} on file {}".format(str(e), filename))
        return 1

class PicoporeRenameRunner(AbstractPicoporeRunner):

    def __init__(self, args):
        super(PicoporeRenameRunner, self).__init__(args)
        self.pattern = args.pattern
        self.replacement = args.replacement
        self.processed = 0

    def get_func(self):
        func = functools.partial(rename, pattern=self.pattern, replacement=self.replacement)
        message = "Renaming {} to {}".format(self.pattern, self.replacement)
        return func, message

    def preprocess(self, fileList):
        self.processed += len(fileList)

    def postprocess(self, results):
        log("Successfully renamed {} of {} files.".format(self.processed - sum(results), self.processed))
        return self.processed

__description = """"A tool for renaming groups and datasets within Oxford Nanopore Technologies FAST5 files"""
        
def main():
    args = parseRenameArgs(description=__description)
    runner = PicoporeRenameRunner(args)
    return runner.execute()

if __name__ == "__main__":
    exit(main())



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
import os
import numpy as np
import glob
import sys
import re

def log(message='', end='\n'):
    print(message, end=end)
    sys.stdout.flush()

def getPrefixedFilename(filename, prefix=""):
    if prefix is None or prefix == "":
        return filename
    elif os.path.isdir(filename):
        return os.path.join(filename, prefix)
    else:
        return os.path.join(os.path.dirname(filename), ".".join([prefix, os.path.basename(filename)]))

def recursiveFindFast5(inp, skip_root=False, depth=0):
    files = []
    for path in inp:
        if os.path.isdir(path):
            files.extend(recursiveFindFast5([os.path.join(path, i) for i in os.listdir(path)], skip_root, depth+1))
        elif (not skip_root or depth > 1) and os.path.isfile(path) and path.endswith(".fast5"):
            files.append(path)
        elif not skip_root:
            files.extend(glob.glob("{}*.fast5".format(path)))
    if depth == 0 and len(files) == 0:
        log("No files found under {}".format(', '.join(inp)))
    return files

def isType(obj, types):
    try:
        return type(obj).__name__ in types
    except TypeError as e:
        if str(e).endswith("is not iterable"):
            # got a single value, not a list
            return type(obj).__name__ == types
        else:
            raise e

def isGroup(obj):
    return isType(obj, ["Group"])

def isDataset(obj):
    return isType(obj, ["Dataset"])

def isInt(obj):
    return isType(obj, ['int', 'int4', 'int8', 'int16', 'int32', 'int64', 'uint', 'uint4', 'uint8', 'uint16', 'uint32', 'uint64'])

def isStr(obj):
    return isType(obj, ['str', 'string_', 'bytes_', 'bytes', 'unicode'])

def isArray(obj):
    return isType(obj, ['list', 'ndarray', 'MaskedArray'])

def isFloat(obj):
    return isType(obj, ['float', 'float16', 'float32', 'float64'])

def getUIntDtype(num):
    if num < 2**8:
        name='uint8'
    elif num < 2**16:
        name='uint16'
    elif num < 2**32:
        name='uint32'
    else:
        name='uint64'
    return name

def getIntDtype(num):
    if abs(num) < 2**7:
        name='int8'
    elif abs(num) < 2**15:
        name='int16'
    elif abs(num) < 2**31:
        name='int32'
    else:
        name='int64'
    return name

def getDtype(data):
    if isArray(data):
        if isInt(data[0]):
            if min(data) > 0:
                name=getUIntDtype(max(data))
            else:
                name=getIntDtype(max(data))
        elif isStr(data[0]):
            name='|S{}'.format(max(max([len(i) for i in data]),1))
        else:
            name=getDtype(data[0])
    elif isInt(data):
        if data > 0:
            name=getUIntDtype(data)
        else:
            name=getIntDtype(data)
    elif isStr(data):
        name='|S{}'.format(max(len(data),1))
    elif isFloat(data):
        # TODO: is there a better way to type floats? sig figs?
        name=type(data).__name__
    else:
        raise TypeError("Data type for value {} not recognised: {}".format(str(data), type(data).__name__))
        return None
    return np.dtype(name)

def recursiveFindDatasets(group, keyword, match_child):
    eventPaths = []
    if isGroup(group):
        for subgroup in group.values():
            eventPaths.extend(recursiveFindDatasets(subgroup, keyword, match_child))
    name = group.name
    if match_child:
        name = name.split("/")[-1]
    if re.search(keyword, name) is not None:
        eventPaths.append(group.name)
    return eventPaths

def findDatasets(f, group_id="all", keyword="Events", entry_point="Analyses", match_child=False):
    eventPaths = []
    try:
        analyses = f.get(entry_point)
        for group in analyses.values():
            if group_id == "all" or group.endswith(group_id):
                eventPaths.extend(recursiveFindDatasets(group, keyword, match_child))
    except AttributeError:
        # no analyses, dont worry
        pass
    return eventPaths

def rewriteDataset(f, path, compression="gzip", compression_opts=1, dataset=None):
    obj = f.get(path)
    if not isDataset(obj):
        return
    attrs = obj.attrs
    dataset = obj.value if dataset is None else dataset
    del f[path]
    try:
        cols = dataset.dtype.names
        if cols is None:
            raise AttributeError("Array dtype is missing names")
        newtype=[(name, getDtype(dataset[name])) for name in dataset.dtype.names]
        f.create_dataset(path, data=dataset.astype(newtype), dtype=newtype, compression=compression, compression_opts=compression_opts)
    except AttributeError:
        try:
            f.create_dataset(path, data=dataset, dtype=getDtype(dataset), compression=compression, compression_opts=compression_opts)
        except TypeError as e:
            if str(e) == "Scalar datasets don't support chunk/filter options":
                f.create_dataset(path, data=dataset, dtype=getDtype(dataset))
            else:
                log(path)
                raise e
    for name, value in attrs.items():
        f[path].attrs[name] = value

def recursiveCollapseGroups(f, basegroup, path, group):
    for subname, object in group.items():
        subpath = "{}.{}".format(path, subname)
        if isGroup(object):
            recursiveCollapseGroups(f, basegroup, subpath, object)
        else:
            f.move(object.name, "{}/{}".format(basegroup, subpath))
    for k, v in group.attrs.items():
        f[basegroup].attrs.create("{}.{}".format(path, k), v, dtype=getDtype(v))
    del f[group.name]

def uncollapseGroups(f, basegroup):
    for name, object in basegroup.items():
        f.move("{}/{}".format(basegroup.name, name), name.replace(".", "/")) # TODO: does this include basegroup?
    for k, v in basegroup.attrs.items():
        k = k.split(".")
        groupname = "/".join(k[:-1])
        attrname = k[-1]
        try:
            f.create_group(groupname)
        except ValueError as e:
            if groupname in f:
                pass
            else:
                raise e
        f[groupname].attrs.create(attrname, v, dtype=getDtype(v))
    del f[basegroup.name]



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

from picopore.parse_args import parseArgs
from picopore.runner import PicoporeCompressionRunner

__description = """See also:
\tpicopore-realtime\tmonitors a directory for new reads and compresses them in real time
\tpicopore-test\t\tcompresses to temporary files and checks that all datasets and attributes are equal (lossless modes only)
\tpicopore-rename\t\trenames groups and datasets within FAST5 files"""

def main():
    args = parseArgs(description=__description)
    runner = PicoporeCompressionRunner(args)
    return runner.execute()

if __name__ == "__main__":
    exit(main())



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys

from argparse import ArgumentParser, ArgumentError, Action, RawDescriptionHelpFormatter
from builtins import input

from picopore.version import __version__

def checkDeprecatedArgs():
    import subprocess
    import warnings
    import signal
    args = sys.argv[1:]
    if "--realtime" in args:
        args.remove('--realtime')
        args = ['picopore-realtime'] + args
        warnings.warn("picopore --realtime will be deprecated in 2.0. Use picopore-realtime instead.",FutureWarning)
    elif "--test" in args:
        args.remove('--test')
        args = ['picopore-test'] + args
        warnings.warn("picopore --test will be deprecated in 2.0. Use picopore-test instead.",FutureWarning)
    else:
        return
    p = subprocess.Popen(args)
    try:
        p.wait()
    except KeyboardInterrupt:
        p.send_signal(signal.SIGINT)
        p.wait()
    exit(p.returncode)

def checkInputs(args):
    args.input = [os.path.abspath(i) for i in args.input]
    # we go recursively - better remove duplicates
    keepDirs = []
    for i in range(len(args.input)):
        subDir = False
        for j in range(len(args.input)):
            if args.input[j] in args.input[i] and i != j:
                subDir = True
        if not subDir:
            keepDirs.append(args.input[i])
    return keepDirs

def checkSure():
    response = input("Are you sure? (yes|no): ")
    assert isinstance(response, str)
    if "yes".startswith(response):
        return 1
    else:
        return 0

class AutoBool(Action):
    def __init__(self, option_strings, dest, default=None, required=False, help=None):
        """Automagically create --foo / --no-foo argument pairs
        Source: https://github.com/nanoporetech/nanonet/blob/master/nanonet/cmdargs.py"""

        if default is None:
            raise ValueError('You must provide a default with AutoBool action')
        if len(option_strings)!=1:
            raise ValueError('Only single argument is allowed with AutoBool action')
        opt = option_strings[0]
        if not opt.startswith('--'):
            raise ValueError('AutoBool arguments must be prefixed with --')

        opt = opt[2:]
        opts = ['--' + opt, '--no-' + opt]
        if default:
            default_opt = opts[0]
        else:
            default_opt = opts[1]
        super(AutoBool, self).__init__(opts, dest, nargs=0, const=None,
                                       default=default, required=required,
                                       help='{} (Default: {})'.format(help, default_opt))
    def __call__(self, parser, namespace, values, option_strings=None):
        if option_strings.startswith('--no-'):
            setattr(namespace, self.dest, False)
        else:
            setattr(namespace, self.dest, True)

def addCommonArgs(parser):
    parser.add_argument('-v', '--version', action='version', version='Picopore {}'.format(__version__), help="show version number and exit")
    parser.add_argument("-y", action="store_true", default=False, help="skip confirm step")
    parser.add_argument("-t", "--threads", type=int, default=1, help="number of threads (Default: 1)", metavar="INT")
    parser.add_argument("--prefix", default=None, help="add prefix to output files to prevent overwrite", metavar="STR")
    parser.add_argument("--skip-root", action=AutoBool, default=False, help="ignore files in root input directories for albacore realtime compression")
    parser.add_argument("--print-every", type=int, default=100, help="print a dot every approximately INT files, or -1 to silence (Default: 100)", metavar="INT")
    parser.add_argument("input", nargs="*", help="list of directories or fast5 files to shrink")
    return parser

__description = """A tool for reducing the size of an Oxford Nanopore Technologies dataset without losing any data"""
def parseArgs(description=None, prog='picopore'):
    checkDeprecatedArgs()
    if description is not None:
        description = __description + "\n\n" + description
    else:
        description = __description
    parser = ArgumentParser(description=description, prog=prog, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("--mode", choices=('lossless', 'deep-lossless', 'raw'), help="choose compression mode", required=True)
    parser.add_argument("--revert", default=False, action="store_true", help="reverts files to original size (lossless modes only)")
    parser.add_argument("--fastq", action=AutoBool, default=True, help="retain FASTQ data (raw mode only)")
    parser.add_argument("--summary", action=AutoBool, default=False, help="retain summary data (raw mode only)")
    parser.add_argument("--manual", default=None, help="manually remove only groups whose paths contain STR (raw mode only, regular expressions permitted, overrides defaults)", metavar="STR")
    parser = addCommonArgs(parser)
    args = parser.parse_args()

    args.input = checkInputs(args)
    args.group = "all" # TODO: is it worth supporting group?

    return args

def parseRenameArgs(description, prog='picopore-rename'):
    parser = ArgumentParser(description=description, prog=prog)
    parser.add_argument('-p', '--pattern', required=True, help="String or regex to replace")
    parser.add_argument('-r', '--replacement', required=True, help="String or regex replacement for PATTERN")
    parser = addCommonArgs(parser)
    args = parser.parse_args()

    return args



"""
    This file is part of Picopore.

    Picopore is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Picopore is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Picopore.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from time import sleep
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from picopore.multiprocess import Multiprocessor
from picopore.runner import PicoporeCompressionRunner
from picopore.compress import chooseCompressFunc
from picopore.parse_args import parseArgs, checkSure
from picopore.util import log

class ReadsFolder(object):
    def __init__(self, runner):
        self.runner = runner

        self.event_handler = PatternMatchingEventHandler(patterns=["*.fast5"],
                ignore_patterns=[],
                ignore_directories=True)
        self.event_handler.on_created = self.on_created
        self.event_handler.on_moved = self.on_moved

        self.observer = Observer()

        self.observedPaths = []
        for path in self.runner.input:
            if os.path.isdir(path):
                self.observer.schedule(self.event_handler, path, recursive=True)
                self.observedPaths.append(path)
        log("Monitoring {} in real time. Press Ctrl+C to exit.".format(", ".join(self.observedPaths)))

    def start(self):
        self.observer.start()
        self.runner.run(postprocess=False)

    def add_file(self, path):
        self.runner.process([path])

    def on_created(self, event):
        if self.runner.skip_root and os.path.dirname(event.src_path) in self.observedPaths:
            return 0
        elif self.runner.prefix is not None and os.path.basename(event.src_path).startswith(self.runner.prefix):
            return 0
        else:
            self.add_file(event.src_path)
            return 0

    def on_moved(self, event):
        if self.runner.skip_root and os.path.dirname(event.src_path) in self.observedPaths and os.path.dirname(event.dest_path) not in self.observedPaths:
            self.add_file(event.dest_path)

    def stop(self):
        log("Processing in-progress files. Press Ctrl-C again to abort.")
        try:
            self.runner.stop()
        except KeyboardInterrupt:
            log("Aborted.")
            pass

        self.observer.stop()
        self.observer.join()

class PicoporeRealtimeRunner(PicoporeCompressionRunner):

    def __init__(self, args):
        super(PicoporeRealtimeRunner, self).__init__(args)
        _, name = chooseCompressFunc(self.revert, self.mode, self.fastq, self.summary, self.manual, realtime=True)
        log(name + "...",end='')
        if self.y:
            log()
        elif checkSure():
            self.y = True
        else:
            exit(1)
        self.readsFolder = ReadsFolder(self)

    def execute(self):
        self.readsFolder.start()
        try:
            while True:
                sleep(5)
        except KeyboardInterrupt:
            log("\nExiting Picopore.")
        self.readsFolder.stop()
        return 0

__description = """"picopore-realtime monitors a directory for new reads and compresses them in real time"""

def main():
    args = parseArgs(description=__description, prog='picopore-realtime')
    runner = PicoporeRealtimeRunner(args)
    return runner.execute()

if __name__ == "__main__":
    exit(main())



from __future__ import print_function
import os
from setuptools import setup
import sys

version_py = os.path.join(os.path.dirname(__file__), 'picopore', 'version.py')
version = open(version_py).read().strip().split('=')[-1].replace('"','').strip()
print(version)

if sys.version_info[:2] < (2,7) or (3,0) <= sys.version_info[:2] < (3,4):
	raise RuntimeError("Python version 2.7 or >=3.4 required.")

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname), 'r') as handle:
        data = handle.read()
    return data

setup(
  name = 'picopore',
  packages = ['picopore'],
  package_dir={'picopore': "picopore"},
  version=version,
  install_requires=['h5py>2.2.0','watchdog','future'],
  description = 'A tool for reducing the size of Oxford Nanopore Technologies\' datasets without losing information.',
  long_description=read('README.rst'),
  author = 'Scott Gigante',
  author_email = 'scottgigante@gmail.com',
  url = 'https://github.com/scottgigante/picopore', # use the URL to the github repo
  keywords = ['nanopore', 'compression', 'hdf5', 'fast5', 'oxford', 'minion'], # arbitrary keywords
  classifiers = [
		"Development Status :: 4 - Beta",
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: GNU General Public License (GPL)',
		'Topic :: Scientific/Engineering :: Bio-Informatics'
		],
  entry_points = {
        'console_scripts': ['picopore = picopore.__main__:main',
                            'picopore-rename = picopore.rename:main',
                            'picopore-test = picopore.test:main',
                            'picopore-realtime = picopore.realtime:main'],
    },
)



from __future__ import absolute_import
from subprocess import call, PIPE
from picopore.util import log

if not call("type h5repack", shell=True, stdout=PIPE, stderr=PIPE) == 0:
    log("h5repack (hdf5-tools) not installed. Aborting.")
    exit(1)
